/* (C) Pavel Surynek, 2009 - 2010  */
/* pavel.surynek@mff.cuni.cz       */
/* graph.h                         */

/*
  Graph representation.
 */


#include "system.h"
#include "version.h"
#include "graph.h"


int Graph::init(int V, int E, int *cN)
{
	m_V = V;

	m_vertices = (Vertex*)malloc(V * sizeof(Vertex));
	if (m_vertices == NULL)
	{
		return -1;
	}
        assert(m_vertices != NULL);

	m_E = E;
	m_cE = 0;

	m_edges = (Edge*)malloc(E * sizeof(Edge));
	if (m_edges == NULL)
	{
		return -1;
	}
	assert(m_edges != NULL);

	m_N = 2 * E;
	m_cN = 0;
	m_neighbors = (Neighbor*)malloc(m_N * sizeof(Neighbor));
	if (m_neighbors == NULL)
	{
		return -1;
	}
	assert(m_neighbors != NULL);

	Neighbor *last_neighbor = m_neighbors;

	for (int i = 0; i < V; ++i)
	{
		m_vertices[i].m_cN = 0;
		m_vertices[i].m_N = cN[i];
		m_vertices[i].m_neighbors = last_neighbor;
		last_neighbor += cN[i];
	}

	return 0;
}


void Graph::add_edge(int u, int v, int weight)
{
	Vertex *vertex_u = &m_vertices[u];
	Vertex *vertex_v = &m_vertices[v];

	Neighbor *neighbor = &vertex_u->m_neighbors[vertex_u->m_cN++];
	assert(vertex_u->m_cN <= vertex_u->m_N);
	Edge *edge = neighbor->m_edge = &m_edges[m_cE++];
	edge->m_weight = weight;
	neighbor->m_vertex = vertex_v;

	neighbor = &vertex_v->m_neighbors[vertex_v->m_cN++];
	assert(vertex_v->m_cN <= vertex_v->m_N);
	neighbor->m_edge = edge;
	neighbor->m_vertex = vertex_u;
}


void Graph::print(void) const
{
	for (int i = 0; i < m_V; ++i)
	{
		printf("%d: {", i);

		Vertex *vertex = &m_vertices[i];

		Neighbor *neighbor = vertex->m_neighbors;
		for (int j = 0; j < vertex->m_cN - 1; ++j)
		{
			printf("%d(%d),", neighbor->m_vertex - m_vertices, neighbor->m_edge->m_weight);
			++neighbor;
		}
		if (vertex->m_cN > 0)
		{
			printf("%d(%d)", neighbor->m_vertex - m_vertices, neighbor->m_edge->m_weight);
		}

		printf("}\n");
	}
}


int IGraph::init(Formula *formula, int V_chunk, int E_chunk, int N_chunk)
{
	m_formula = formula;

	m_last_id = 0;

	m_E = m_N = 0;
	m_cV = m_cE = m_cN = 0;
	m_ccV = m_ccE = m_ccN = 0;

	m_V_chunk = V_chunk;
	m_E_chunk = E_chunk;
	m_N_chunk = N_chunk;

	m_first_vchunk = (IVertexChunk*)malloc(sizeof(IVertexChunk));
	if (m_first_vchunk == NULL)
	{
		return -1;
	}
	assert(m_first_vchunk != NULL);
	
	m_first_vchunk->m_vertices = (IVertex*)malloc(V_chunk * sizeof(IVertex));
	if (m_first_vchunk->m_vertices == NULL)
	{
		return -1;
	}
	assert(m_first_vchunk->m_vertices != NULL);
	m_first_vchunk->m_next = NULL;
	m_V = m_V_chunk;

	m_first_echunk = (IEdgeChunk*)malloc(sizeof(IEdgeChunk));
	if (m_first_echunk == NULL)
	{
		return -1;
	}
	assert(m_first_echunk != NULL);
	
	m_first_echunk->m_edges = (IEdge*)malloc(E_chunk * sizeof(IEdge));
	if (m_first_echunk->m_edges == NULL)
	{
		return -1;
	}
	assert(m_first_echunk->m_edges != NULL);
	m_first_echunk->m_next = NULL;
	m_E = m_E_chunk;

	m_first_nchunk = (INeighborChunk*)malloc(sizeof(INeighborChunk));
	if (m_first_nchunk == NULL)
	{
		return -1;
	}
	assert(m_first_nchunk != NULL);
	
	m_first_nchunk->m_nnodes = (INeighborNode*)malloc(N_chunk * sizeof(INeighborNode));
	if (m_first_nchunk->m_nnodes == NULL)
	{
		return -1;
	}
	assert(m_first_nchunk->m_nnodes != NULL);
	m_first_nchunk->m_next = NULL;
	m_N = m_N_chunk;

	m_last_vchunk = &m_first_vchunk;
	m_last_echunk = &m_first_echunk;
	m_last_nchunk = &m_first_nchunk;

	return 0;
}


int IGraph::add_vertex(int cV)
{
	int result;

	IVertexChunk *last_vchunk = *m_last_vchunk;

	if ((result = alloc_vertices(cV)) < 0)
	{
		return result;
	}

	for (IVertexChunk *vchunk = last_vchunk; vchunk != NULL; vchunk = vchunk->m_next)
	{
		int fill_V = m_V_chunk - m_ccV;
		if (cV > fill_V)
		{
			while (m_ccV < m_V_chunk)
			{
				IVertex *vertex = &vchunk->m_vertices[m_ccV++];
				vertex->m_id = m_last_id++;
				vertex->m_cN = 0;
				vertex->m_first_nnode = NULL;
				vertex->m_last_nnode = &vertex->m_first_nnode;
			}
			cV -= fill_V;
			m_cV += fill_V;
			m_ccV = 0;
		}
		else
		{
			m_cV += cV;

			while (cV > 0)
			{
				IVertex *vertex = &vchunk->m_vertices[m_ccV++];
				vertex->m_id = m_last_id++;
				vertex->m_cN = 0;
				vertex->m_first_nnode = NULL;
				vertex->m_last_nnode = &vertex->m_first_nnode;
				--cV;
			}
		}
	}

	return 0;
}


int IGraph::add_edge(int u, int v, int weight)
{
	int result;

	IEdge *edge;
	INeighborNode *nnode_u, *nnode_v;

	if ((result = alloc_edge(&edge)) < 0)
	{
		return result;
	}
	if ((result = alloc_neighbor(&nnode_u)) < 0)
	{
		return result;
	}
	if ((result = alloc_neighbor(&nnode_v)) < 0)
	{
		return result;
	}

	IVertex *vertex_u = get_vertex(u);
	IVertex *vertex_v = get_vertex(v);

	edge->m_weight = weight;

	nnode_u->m_edge = edge;
	nnode_u->m_vertex = vertex_v;
	nnode_u->m_next = NULL;

	*vertex_u->m_last_nnode = nnode_u;
	vertex_u->m_last_nnode = &nnode_u->m_next;
	++vertex_u->m_cN;

	nnode_v->m_edge = edge;
	nnode_v->m_vertex = vertex_u;
	nnode_v->m_next = NULL;

	*vertex_v->m_last_nnode = nnode_v;
	vertex_v->m_last_nnode = &nnode_v->m_next;
	++vertex_v->m_cN;

	return 0;
}


int IGraph::increase_edge_weight(int u, int v, int dweight)
{
	int result;

	IEdge *edge = get_edge(u, v);

	if (edge != NULL)
	{
		edge->m_weight += dweight;
	}
	else
	{
		if ((result = add_edge(u, v, dweight)) < 0)
		{
			return result;
		}
	}

	return 0;
}


IVertex* IGraph::get_vertex(int id)
{
	IVertexChunk *vchunk = m_first_vchunk;

	while (id >= m_V_chunk)
	{
		vchunk = vchunk->m_next;
		id -= m_V_chunk;
	}

	return &vchunk->m_vertices[id];
}


IEdge* IGraph::get_edge(int u, int v)
{
	IVertex *vertex_u = get_vertex(u);
	IVertex *vertex_v = get_vertex(v);

	if (vertex_u->m_cN > vertex_v->m_cN)
	{
		INeighborNode *nnode = vertex_v->m_first_nnode;

		while (nnode != NULL)
		{
			if (nnode->m_vertex->m_id == u)
			{
				return nnode->m_edge;
			}
			nnode = nnode->m_next;
		}
		return NULL;
	}
	else
	{
		INeighborNode *nnode = vertex_u->m_first_nnode;

		while (nnode != NULL)
		{
			if (nnode->m_vertex->m_id == v)
			{
				return nnode->m_edge;
			}
			nnode = nnode->m_next;
		}
		return NULL;
	}
}


void IGraph::print(void) const
{
/*
	printf("m_V:%d, m_cV:%d, m_ccV:%d\n", m_V, m_cV, m_ccV);
	printf("m_E:%d, m_cE:%d, m_ccE:%d\n", m_E, m_cE, m_ccE);
	printf("m_N:%d, m_cN:%d, m_ccN:%d\n", m_N, m_cN, m_ccN);
*/

	int cV = 0;

	for (IVertexChunk *vchunk = m_first_vchunk; vchunk != NULL; vchunk = vchunk->m_next)
	{
		if (m_cV - cV > m_V_chunk)
		{
			for (int i = 0; i < m_V_chunk; ++i)
			{
				IVertex *vertex = &vchunk->m_vertices[i];
				printf("%d: {", vertex->m_id);

				INeighborNode *nnode = vertex->m_first_nnode;
				for (int j = 0; j < vertex->m_cN - 1; ++j)
				{
					printf("%d(%d),", nnode->m_vertex->m_id, nnode->m_edge->m_weight);
					nnode = nnode->m_next;
				}
				if (vertex->m_cN > 0)
				{
					printf("%d(%d)", nnode->m_vertex->m_id, nnode->m_edge->m_weight);
				}
				printf("}\n");
			}
			cV += m_V_chunk;
		}
		else
		{
			for (int i = 0; i < m_ccV; ++i)
			{
				IVertex *vertex = &vchunk->m_vertices[i];
				printf("%d: {", vertex->m_id);

				INeighborNode *nnode = vertex->m_first_nnode;
				for (int j = 0; j < vertex->m_cN - 1; ++j)
				{
					printf("%d(%d),", nnode->m_vertex->m_id, nnode->m_edge->m_weight);
					nnode = nnode->m_next;
				}
				if (vertex->m_cN > 0)
				{
					printf("%d(%d)", nnode->m_vertex->m_id, nnode->m_edge->m_weight);
				}
				printf("}\n");
			}
			break;
		}
	}
}


void IGraph::export_GR(bool exp_literals, bool exp_clusters, bool exp_layers, bool exp_visits) const
{
	char label[128], label_neighbor[128], data[128];

	printf("<graph>\n");
	printf("    <nodes>\n");

	int cV = 0;

	for (IVertexChunk *vchunk = m_first_vchunk; vchunk != NULL; vchunk = vchunk->m_next)
	{
		if (m_cV - cV > m_V_chunk)
		{
			for (int i = 0; i < m_V_chunk; ++i)
			{
				IVertex *vertex = &vchunk->m_vertices[i];

				get_label(label, vertex, exp_literals);
				get_data(data, vertex, exp_clusters, exp_layers, exp_visits);

				printf("        <node label=\"%s\" data=\"%s\">\n", label, data);
				printf("            <position>\n");
				printf("                <2D_point x='0' y='0'>\n");
				printf("            </position>\n");
				printf("        </node>\n");
			}
			cV += m_V_chunk;
		}
		else
		{
			for (int i = 0; i < m_ccV; ++i)
			{
				IVertex *vertex = &vchunk->m_vertices[i];

				get_label(label, vertex, exp_literals);
				get_data(data, vertex, exp_clusters, exp_layers, exp_visits);

				printf("        <node label=\"%s\" data=\"%s\">\n", label, data);
				printf("            <position>\n");
				printf("                <2D_point x='0' y='0'>\n");
				printf("            </position>\n");
				printf("        </node>\n");

			}
			break;
		}
	}
	printf("    </nodes>\n");
	printf("    <edges>\n");

	cV = 0;

	for (IVertexChunk *vchunk = m_first_vchunk; vchunk != NULL; vchunk = vchunk->m_next)
	{
		if (m_cV - cV > m_V_chunk)
		{
			for (int i = 0; i < m_V_chunk; ++i)
			{
				IVertex *vertex = &vchunk->m_vertices[i];

				INeighborNode *nnode = vertex->m_first_nnode;
				for (int j = 0; j < vertex->m_cN; ++j)
				{
					if (nnode->m_vertex->m_id > vertex->m_id)
					{
						get_label(label_neighbor, nnode->m_vertex, exp_literals);
						get_label(label, vertex, exp_literals);

						printf("        <edge>\n");
						printf("            <first_node>%s</first_node>\n", label_neighbor);
						printf("            <second_node>%s</second_node>\n", label);
						printf("        </edge>\n");
					}
					nnode = nnode->m_next;
				}
			}
			cV += m_V_chunk;
		}
		else
		{
			for (int i = 0; i < m_ccV; ++i)
			{
				IVertex *vertex = &vchunk->m_vertices[i];

				INeighborNode *nnode = vertex->m_first_nnode;
				for (int j = 0; j < vertex->m_cN; ++j)
				{
					if (nnode->m_vertex->m_id > vertex->m_id)
					{
						get_label(label_neighbor, nnode->m_vertex, exp_literals);
						get_label(label, vertex, exp_literals);

						printf("        <edge>\n");
						printf("            <first_node>%s</first_node>\n", label_neighbor);
						printf("            <second_node>%s</second_node>\n", label);
						printf("        </edge>\n");
					}
					nnode = nnode->m_next;
				}
			}
			break;
		}
	}
	printf("    </edges>\n");
	printf("</graph>\n");
}


void IGraph::get_label(char *label, IVertex *vertex, bool exp_literals) const
{
	if (exp_literals)
	{
		char literal_str[32];
		m_formula->sprint_literal(literal_str, vertex->m_literal);
		sprintf(label, "%d (%s)", vertex->m_id, literal_str);
	}
	else
	{
		sprintf(label, "%d", vertex->m_id);
	}
}


void IGraph::get_data(char *data, IVertex *vertex, bool exp_clusters, bool exp_layers, bool exp_visits) const
{
	if (exp_clusters)
	{
		if (exp_layers)
		{
			if (exp_visits)
			{
				sprintf(data, "c=%d l=%d v=%d", vertex->m_cluster, vertex->m_layer, vertex->m_visits);
			}
			else
			{
				sprintf(data, "c=%d l=%d", vertex->m_cluster, vertex->m_layer);
			}
		}
		else
		{
			if (exp_visits)
			{
				sprintf(data, "c=%d v=%d", vertex->m_cluster, vertex->m_visits);
			}
			else
			{
				sprintf(data, "c=%d", vertex->m_cluster);
			}
		}
	}
	else
	{
		if (exp_layers)
		{
			if (exp_visits)
			{
				sprintf(data, "l=%d v=%d", vertex->m_layer, vertex->m_visits);
			}
			else
			{
				sprintf(data, "l=%d", vertex->m_layer);
			}
		}
		else
		{
			if (exp_visits)
			{
				sprintf(data, "v=%d", vertex->m_visits);
			}
			else
			{
				// nothing
			}
		}
	}
}


int IGraph::alloc_vertices(int cV)
{
	int required = m_cV + cV;

	while (required > m_V)
	{
		m_last_vchunk = &(*m_last_vchunk)->m_next;

		*m_last_vchunk = (IVertexChunk*)malloc(sizeof(IVertexChunk));
		if (*m_last_vchunk == NULL)
		{
			return -1;
		}
		assert(*m_last_vchunk != NULL);

		(*m_last_vchunk)->m_vertices = (IVertex*)malloc(m_V_chunk * sizeof(IVertex));
		if ((*m_last_vchunk)->m_vertices == NULL)
		{
			return -1;
		}
		assert((*m_last_vchunk)->m_vertices != NULL);

		(*m_last_vchunk)->m_next = NULL;
		m_V += m_V_chunk;
	}

	return 0;
}


int IGraph::alloc_edge(IEdge **edge)
{
	if (m_cE + 1 > m_E)
	{
		m_last_echunk = &(*m_last_echunk)->m_next;

		*m_last_echunk = (IEdgeChunk*)malloc(sizeof(IEdgeChunk));
		if (*m_last_echunk == NULL)
		{
			return -1;
		}
		assert(*m_last_echunk != NULL);

		(*m_last_echunk)->m_edges = (IEdge*)malloc(m_E_chunk * sizeof(IEdge));
		if ((*m_last_echunk)->m_edges == NULL)
		{
			return -1;
		}
		assert((*m_last_echunk)->m_edges != NULL);

		(*m_last_echunk)->m_next = NULL;
		m_E += m_E_chunk;

		m_ccE = 0;
	}

	*edge = &(*m_last_echunk)->m_edges[m_ccE++];
	++m_cE;

	return 0;
}


int IGraph::alloc_neighbor(INeighborNode **nnode)
{
	if (m_cN + 1 > m_N)
	{
		m_last_nchunk = &(*m_last_nchunk)->m_next;

		*m_last_nchunk = (INeighborChunk*)malloc(sizeof(INeighborChunk));
		if (*m_last_nchunk == NULL)
		{
			return -1;
		}
		assert(*m_last_nchunk != NULL);

		(*m_last_nchunk)->m_nnodes = (INeighborNode*)malloc(m_N_chunk * sizeof(INeighborNode));
		if ((*m_last_nchunk)->m_nnodes == NULL)
		{
			return -1;
		}
		assert((*m_last_nchunk)->m_nnodes != NULL);

		(*m_last_nchunk)->m_next = NULL;
		m_N += m_N_chunk;

		m_ccN = 0;
	}

	*nnode = &(*m_last_nchunk)->m_nnodes[m_ccN++];
	++m_cN;

	return 0;
}


int Clustering::init(const Formula *formula, Clause **path, int length, IGraph *graph)
{
	m_formula = formula;
	m_graph = graph;
	m_cC = length;

	m_clusters = (Cluster*)malloc(length * sizeof(Cluster));
	if (m_clusters == NULL)
	{
		return -1;
	}
	assert(m_clusters != NULL);
	
	for (int i = 0; i < length; ++i)
	{
		Cluster *cluster = &m_clusters[i];
		cluster->m_cV = path[i]->m_cL;
		cluster->m_vertices = (IVertex**)malloc(path[i]->m_cL * sizeof(IVertex*));

		if (cluster->m_vertices == NULL)
		{
			return -1;
		}
		assert(cluster->m_vertices != NULL);
	}

	return 0;
}


void Clustering::print(void) const
{
	printf("Clustering:\n");
	for (int i = 0; i < m_cC; ++i)
	{
		Cluster *cluster = &m_clusters[i];
		printf("%d: ", i);

		int j;

		for (j = 0; j < cluster->m_cV - 1; ++j)
		{
			printf("%d(", cluster->m_vertices[j]->m_id);
			m_formula->print_literal(cluster->m_vertices[j]->m_literal);
			printf("),");
		}
		if (cluster->m_cV > 0)
		{
			printf("%d(", cluster->m_vertices[j]->m_id);
			m_formula->print_literal(cluster->m_vertices[j]->m_literal);
			printf(")");
		}
		printf("\n");
	}
}


void Clustering::print_distances(void) const
{
	printf("Clustering:\n");
	for (int i = 0; i < m_cC; ++i)
	{
		Cluster *cluster = &m_clusters[i];
		printf("%d: ", i);

		int j;

		for (j = 0; j < cluster->m_cV - 1; ++j)
		{
			printf("%d(", cluster->m_vertices[j]->m_id);
			m_formula->print_literal(cluster->m_vertices[j]->m_literal);
			printf(",%d),", cluster->m_vertices[j]->m_distance);
		}
		if (cluster->m_cV > 0)
		{
			printf("%d(", cluster->m_vertices[j]->m_id);
			m_formula->print_literal(cluster->m_vertices[j]->m_literal);
			printf(",%d)", cluster->m_vertices[j]->m_distance);
		}
		printf("\n");
	}
}


void Clustering::print_visit_matrices(void) const
{
	printf("Clustering:\n");
	for (int i = 0; i < m_cC; ++i)
	{
		Cluster *cluster = &m_clusters[i];
		printf("%d:\n", i);

		int j;

		for (j = 0; j < cluster->m_cV; ++j)
		{
			printf("%d(", cluster->m_vertices[j]->m_id);
			m_formula->print_literal(cluster->m_vertices[j]->m_literal);
			printf(")\n");
			cluster->m_vertices[j]->m_vmatrix.print();
		}
		printf("\n");
	}
}


int Layering::init(const Formula *formula, Clause **path, int length, IGraph *graph)
{
	m_formula = formula;
	m_graph = graph;

	m_L = m_cL = 0;
	for (int l = 0; l < length; ++l)
	{
		if (path[l]->m_cL > m_L)
		{
			m_L = path[l]->m_cL;
		}
	}

	m_layers = (Layer*)malloc(m_L * sizeof(Layer));
	if (m_layers == NULL)
	{
		return -1;
	}
	assert(m_layers != NULL);

	for (int i = 0; i < m_L; ++i)
	{
		Layer *layer = &m_layers[i];
		layer->m_V = length;
		layer->m_cV = 0;

		layer->m_vertices = (IVertex**)malloc(length * sizeof(IVertex*));
		if (layer->m_vertices == NULL)
		{
			return -1;
		}
		assert(layer->m_vertices != NULL);
	}

	return 0;
}


void Layering::print(void) const
{
	printf("Layering:\n");
	for (int i = 0; i < m_cL; ++i)
	{
		Layer *layer = &m_layers[i];
		printf("%d: ", i);

		int j;

		for (j = 0; j < layer->m_cV - 1; ++j)
		{
			printf("%d (", layer->m_vertices[j]->m_id);
			m_formula->print_literal(layer->m_vertices[j]->m_literal);
			printf("), ");
		}
		if (layer->m_cV > 0)
		{
			printf("%d (", layer->m_vertices[j]->m_id);
			m_formula->print_literal(layer->m_vertices[j]->m_literal);
			printf(")");
		}
		printf("\n");
	}
}


void Layering::print_visits(void) const
{
	printf("Layering:\n");
	for (int i = 0; i < m_cL; ++i)
	{
		Layer *layer = &m_layers[i];
		printf("%d: ", i);

		int j;

		for (j = 0; j < layer->m_cV - 1; ++j)
		{
			printf("%d (", layer->m_vertices[j]->m_id);
			m_formula->print_literal(layer->m_vertices[j]->m_literal);
			printf(")-%d, ", layer->m_vertices[j]->m_visits);
		}
		if (layer->m_cV > 0)
		{
			printf("%d (", layer->m_vertices[j]->m_id);
			m_formula->print_literal(layer->m_vertices[j]->m_literal);
			printf(")-%d", layer->m_vertices[j]->m_visits);
		}
		printf("\n");
	}
}


int Layering::init_visit_matrices(int L, int max_visits)
{
	int result;

	for (int i = 0; i < m_cL; ++i)
	{
		Layer &layer = m_layers[i];

		for (int j = 0; j < layer.m_cV; ++j)
		{
			IVertex *vertex = layer.m_vertices[j];
			if ((result = vertex->m_vmatrix.init(L, max_visits)) < 0)
			{
				return result;
			}
		}
	}

	return 0;
}


int VMatrix::init(int L, int max_visits)
{
	m_L = L;
	m_max_visits = max_visits;

	m_cPaths = (int**)malloc(L * sizeof(int*));
	if (m_cPaths == NULL)
	{
		return -1;
	}
	assert(m_cPaths != NULL);

	for (int i = 0; i < L; ++i)
	{
		m_cPaths[i] = (int*)malloc(max_visits * sizeof(int));
		if (m_cPaths[i] == NULL)
		{
			return -1;
		}
		assert(m_cPaths[i] != NULL);
	}

	for (int i = 0; i < L; ++i)
	{
		for (int j = 0; j < max_visits; ++j)
		{
			m_cPaths[i][j] = 0;
		}
	}

	return 0;
}


void VMatrix::print(void)
{
	printf("VMatrix\n");

	for (int i = 0; i < m_L; ++i)
	{
		for (int j = 0; j < m_max_visits; ++j)
		{
			printf("%d\t", m_cPaths[i][j]);
		}
		printf("\n");
	}
}
