/* (C) Pavel Surynek, 2009 - 2010  */
/* pavel.surynek@mff.cuni.cz       */
/* system.cpp                      */

/*
  Systematic module of sigmaSAT solver.
 */


#include "system.h"
#include "version.h"


int STAT_DECISIONS = 0;
int STAT_PROPAGATIONS = 0;
int STAT_BACKTRACKS = 0;

int EDGE_CHUNK = 1024;
int NEIGHBOR_CHUNK = 1024;


int Formula::build_conflict_graph(IGraph *graph)
{
	graph->init(this, m_cC, EDGE_CHUNK, NEIGHBOR_CHUNK);
	graph->add_vertex(m_cC);

	for (int i = 0; i < m_cC; ++i)
	{
		Clause *origin = &m_clauses[i];
		Literal **literal = origin->m_literals;

		while (*literal != NULL)
		{
			set_TRUE_propagate(*literal);

			Literal **chng_literal = m_chng_literals;
			
			while (chng_literal < m_last_chng)
			{
				ONode *onode = (*chng_literal)->m_cmpl->m_occs;

				while (onode != NULL)
				{
					graph->increase_edge_weight(origin - m_clauses, onode->m_clause - m_clauses, 1);
					onode = onode->m_next;
				}
				++chng_literal;
			}
			unset_unpropagate();
			++literal;
		}
	}
}


int Formula::build_conflict_path(Clause **path, int length, IGraph *graph, Clustering *clustering, Layering *layering)
{
	int result;

	if ((result = clustering->init(this, path, length, graph)) < 0)
	{
		return result;
	}

	int V = 0;

	for (int l = 0; l < length; ++l)
	{
		V += path[l]->m_cL;
	}

	graph->init(this, V, EDGE_CHUNK, NEIGHBOR_CHUNK);
	graph->add_vertex(V);

	graph->m_cVO = 0;

	graph->m_vo_nodes = (VONode*)malloc(V * sizeof(VONode));
	if (graph->m_vo_nodes == NULL)
	{
		return -1;
	}
	assert(graph->m_vo_nodes != NULL);

	int id = 0;

	for (int l = 0; l < length; ++l)
	{
		Literal **literal = path[l]->m_literals;
		IVertex **cvertex = clustering->m_clusters[l].m_vertices;

		while (*literal != NULL)
		{
			IVertex *vertex = graph->get_vertex(id++);
			vertex->m_literal = *literal++;
			vertex->m_cluster = l;
			vertex->m_layer = -1;
			(*cvertex++) = vertex;

			VONode &vo_node = graph->m_vo_nodes[graph->m_cVO++];
			vo_node.m_next = vertex->m_literal->m_voccs;
			vo_node.m_vertex = vertex;

			vertex->m_literal->m_voccs = &vo_node;
		}
	}

	for (int l = 0; l < length - 1; ++l)
	{
		Cluster &cluster = clustering->m_clusters[l];

		for (int j = 0; j < cluster.m_cV; ++j)
		{
			IVertex *vertex = cluster.m_vertices[j];
			Literal *literal = vertex->m_literal;

			set_TRUE_propagate(literal);
			
			Literal **chng_literal = m_chng_literals;
			
			while (chng_literal < m_last_chng)
			{
				VONode *vo_node = (*chng_literal)->m_cmpl->m_voccs;
				
				while (vo_node != NULL)
				{
					if (vo_node->m_vertex)
					{
						if (vertex->m_id < vo_node->m_vertex->m_id && vertex->m_cluster != vo_node->m_vertex->m_cluster)
						{
							graph->add_edge(vertex->m_id, vo_node->m_vertex->m_id);
						}
					}
					vo_node = vo_node->m_next;
				}
				++chng_literal;
			}
			unset_unpropagate();
		}
	}

	if ((result = layering->init(this, path, length, graph)) < 0)
	{
		return result;
	}

	for (int i = 0; i < clustering->m_cC; ++i)
	{
		Cluster &cluster = clustering->m_clusters[i];

		for (int j = 0; j < cluster.m_cV; ++j)
		{
			if (cluster.m_vertices[j]->m_layer == -1)
			{
				build_next_layer(i, j, graph, clustering, layering);
			}
		}
	}

	return 0;
}


void Formula::build_next_layer(int cluster_idx, int vertex_idx, IGraph *graph, Clustering *clustering, Layering *layering)
{
	Layer &layer = layering->m_layers[layering->m_cL];
	IVertex *vertex = clustering->m_clusters[cluster_idx].m_vertices[vertex_idx];
	vertex->m_layer = layering->m_cL;
	layer.m_vertices[layer.m_cV++] = vertex;

	for (int c = cluster_idx + 1; c < clustering->m_cC; ++c)
	{
		Cluster &cluster = clustering->m_clusters[c];

		int best_back = 0;
		IVertex *best_vertex = NULL;

		for (int v = 0; v < cluster.m_cV; ++v)
		{
			vertex = cluster.m_vertices[v];

			if (vertex->m_layer == -1)
			{
				int back = 0;
				for (int i = 0; i < layer.m_cV; ++i)
				{
					if (graph->get_edge(layer.m_vertices[i]->m_id, vertex->m_id) != NULL)
					{
						++back;
					}
				}
				if (best_vertex == NULL || best_back < back)
				{
					best_vertex = vertex;
					best_back = back;
				}
			}
		}

		if (best_vertex != NULL)
		{
			layer.m_vertices[layer.m_cV++] = best_vertex;
			best_vertex->m_layer = layering->m_cL;
		}
	}
	++layering->m_cL;
}


void Formula::calculate_maximum_visits(Layering *layering)
{
	for (int i = 0; i < layering->m_cL; ++i)
	{
		Layer &layer = layering->m_layers[i];
		int visits_1 = 0;

		for (int j = 0; j < layer.m_cV; ++j)
		{
			IVertex *vertex = layer.m_vertices[j];
			int max_prev = 0;

			for (int k = 0; k < j; ++k)
			{
				IVertex *prev_vertex = layer.m_vertices[k];

				if (layering->m_graph->get_edge(prev_vertex->m_id, vertex->m_id) == NULL)
				{
					if (prev_vertex->m_visits > max_prev)
					{
						max_prev = prev_vertex->m_visits;
					}
				}
			}
			int visits_2 = max_prev + 1;

			if (visits_1 > visits_2)
			{
				vertex->m_visits = visits_1;
			}
			else
			{
				vertex->m_visits = visits_2;
			}
			visits_1 = vertex->m_visits;
		}
	}
}


void Formula::check_paths_PC(Clustering *clustering, Layering *layering)
{
	Cluster &first_cluster = clustering->m_clusters[0];
	Cluster &last_cluster = clustering->m_clusters[clustering->m_cC - 1];

	for (int v1 = 0; v1 < first_cluster.m_cV; ++v1)
	{
		for (int v2 = 0; v2 < last_cluster.m_cV; ++v2)
		{
			IVertex *vertex_1 = first_cluster.m_vertices[v1];
			IVertex *vertex_2 = last_cluster.m_vertices[v2];

			bool exists = check_path_PC(vertex_1, vertex_2, clustering, layering);

			if (exists)
			{
				printf("+(%d,%d)\n", vertex_1->m_id, vertex_2->m_id);
			}
			else
			{
				printf("-(%d,%d)\n", vertex_1->m_id, vertex_2->m_id);
			}
		}
	}
}


bool Formula::check_path_PC(IVertex *first_vertex, IVertex *last_vertex, Clustering *clustering, Layering *layering)
{
	Cluster &first_cluster = clustering->m_clusters[0];

	last_vertex->m_distance = -1;

	for (int v = 0; v < first_cluster.m_cV; ++v)
	{
		IVertex *vertex = first_cluster.m_vertices[v];
		vertex->m_distance = -1;
	}
	first_vertex->m_distance = 0;

	for (int c = 0; c < clustering->m_cC - 1; ++c)
	{
		Cluster &cluster_1 = clustering->m_clusters[c];
		Cluster &cluster_2 = clustering->m_clusters[c + 1];

		for (int v2 = 0; v2 < cluster_2.m_cV; ++v2)
		{
			IVertex *vertex_2 = cluster_2.m_vertices[v2];
			vertex_2->m_distance = -1;
			
		}

		for (int v1 = 0; v1 < cluster_1.m_cV; ++v1)
		{
			IVertex *vertex_1 = cluster_1.m_vertices[v1];

			for (int v2 = 0; v2 < cluster_2.m_cV; ++v2)
			{
				IVertex *vertex_2 = cluster_2.m_vertices[v2];

				if (vertex_1->m_distance >= 0)
				{
					if (clustering->m_graph->get_edge(vertex_1->m_id, vertex_2->m_id) == NULL)
					{
						vertex_2->m_distance = vertex_1->m_distance + 1;
					}
				}
			}
		}
	}

	if (last_vertex->m_distance >= 0)
	{
		return true;
	}
	else
	{
		return false;
	}
}


void insert_watch(WNode *head, WNode *insert)
{
	WNode *next = head->m_next;

	next->m_prev = insert;
	head->m_next = insert;

	insert->m_prev = head;
	insert->m_next = next;
}


void remove_watch(WNode *UNUSED(head), WNode *remove)
{
	WNode *next = remove->m_next;
	WNode *prev = remove->m_prev;

	prev->m_next = next;
	next->m_prev = prev;
}


void skip_comment_DIMACS(FILE *fr)
{
	while (fgetc(fr) != '\n');
}


void swap_affects(Clause *clause_1, Clause *clause_2)
{
	Clause *temp, **tmp;

	temp = *clause_1->m_sort;
	*clause_1->m_sort = *clause_2->m_sort;
	*clause_2->m_sort = temp;

	tmp = clause_1->m_sort;
	clause_1->m_sort = clause_2->m_sort;
	clause_2->m_sort = tmp;
}

enum COMMAND
{
	UNDEFINED = 0,
	SOLVE     = 1
};


void print_intro(void)
{
	printf("sigmaSAT: systemSIGMA %s\n", VERSION);
	printf("Copyright (C) 2009 - 2010 Pavel Surynek\n");
	printf("---------------------------------------\n");
}


int main(int argc, char **argv)
{
	int result;
	COMMAND cmd;
	char *filename;

	clock_t begin, end;

	begin = clock();

//	print_intro();

	if (argc == 2)
	{
		cmd = SOLVE;
		filename = argv[1];
	}
	else
	{
		cmd = UNDEFINED;
	}


	switch (cmd)
	{
	case SOLVE:
	{
		Formula formula;
		formula.load_DIMACS(filename);

/*
		IGraph conflict_graph;
		formula.build_conflict_graph(&conflict_graph);
		conflict_graph.export_GR();
		printf("\n");
*/

		int length;
		const int MAX_LENGTH = 8;
		Clause *path[MAX_LENGTH];

		for (int i = 0; i < formula.m_cC; ++i)
		{
//			printf("Conflicting path for clause:%d\n", i);
			length = formula.find_conflicting_path(&formula.m_clauses[i], path, MAX_LENGTH);
//			formula.print_path_verbose(path, length);

			IGraph graph;
			Clustering clustering;
			Layering layering;

			result = formula.build_conflict_path(path, length, &graph, &clustering, &layering);
			if (result < 0)
			{
				printf("Unable to build conflict path.\n");
				return result;
			}
//			layering.print();
			formula.calculate_maximum_visits(&layering);
//			layering.print_visits();
/*
			graph.print();
			clustering.print();
*/
/*
			formula.check_paths_PC(&clustering, &layering);
			clustering.print_distances();

			layering.init_visit_matrices(layering.m_cL, clustering.m_cC);
			clustering.print_visit_matrices();
*/
			
			graph.export_GR(true, true, true, true);

			break;
		}

//		formula.print();

/*
		bool solved = formula.solve_1();

		end = clock();

		if (solved)
		{
			printf("SAT\n");
		}
		else
		{
			printf("UNSAT\n");
		}
*/

#ifdef STAT
/*
		printf("Decisions:%d\n", STAT_DECISIONS);
		printf("Backtracks:%d\n", STAT_BACKTRACKS);
		printf("Propagations:%d\n", STAT_PROPAGATIONS);
*/
#endif
//		printf("Time:%.3f\n", (end - begin) / (double)CLOCKS_PER_SEC);
		break;
	}
	default:
	{
		printf("No command specified.\n");
		break;
	}
	}

	return 0;
}
